

#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>

void RowMatrixVectorMultiply(int n, double *a, double *b, double *x,
			     MPI_Comm comm)
{
  int i, j;
  int nlocal;        /* Number of locally stored rows of A */
  double *fb;        /* Will point to a buffer that stores the entire vector b */
  int npes, myrank;
  
  int *recvcount;    /* How many rows each process holds */
  int *displs;       /* How many rows held by the previous processes */
  
  MPI_Status status;

  /* Get information about the communicator */
  MPI_Comm_size(comm, &npes);
  MPI_Comm_rank(comm, &myrank);

  /* Allocate the memory that will store the entire vector b */
  fb = (double *)malloc(n*sizeof(double));

  /* Add 1 row to n%npes workers */
  nlocal = n/npes+(n%npes>myrank); 

  /* Gather information for Allgatherv */
  recvcount = malloc(npes*sizeof(int));
  displs    = malloc(npes*sizeof(int));

  /* make arrays containing information about number to receive
     and where in the final array the information will be stored */
  int c = 0;
  for (i=0; i<npes; i++) {
    recvcount[i] = n/npes+(n%npes>i);
    displs[i] = c;
    c += recvcount[i];
  }

  /* Gather the entire vector b on each processor using MPI's ALLGATHERV operation */
  MPI_Allgatherv(b, nlocal, MPI_DOUBLE, fb, recvcount, displs, MPI_DOUBLE, comm);

  /* Perform the matrix-vector multiplication involving the locally stored submatrix */
  for (i=0; i<nlocal; i++) {
    x[i] = 0.0;
    for (j=0; j<n; j++)
      x[i] += a[i*n+j]*fb[j];
  }

  for(i=0; i <nlocal; i++) {
    printf("%d %.1f\n", myrank, x[i]);
  }

  free(fb);
}

int main (int argc, char * argv[]) {
  int i, j;
  int npes, myrank;
  int n, nlocal;
  double *x, *b, *a;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &npes);
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

  if (myrank == 0) {
    if (argc < 2) {
      fprintf(stderr, "usage: %s n\n", argv[0]);
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }

    n = atoi(argv[1]);
  }

  MPI_Bcast(&n, 1, MPI_INT, 0, MPI_COMM_WORLD);

  nlocal = n/npes+(n%npes>myrank);

  b = (double *)malloc(nlocal*sizeof(double));

  for (i=0; i<nlocal; i++) {
    b[i] = 1.0;         // some vector b
  }

  a = (double *)malloc(n*nlocal*sizeof(double));

  for (i=0; i<nlocal; i++) {
    for (j=0; j<n; j++) {
      a[i*n+j] = 0;
      if(i==j)
	a[i*n+j] = 1;    // creating the I matrix
    }
  }

  x = (double *)malloc(nlocal*sizeof(double));

  RowMatrixVectorMultiply(n, a, b, x, MPI_COMM_WORLD);

  free(x);
  free(a);
  free(b);
  
  MPI_Finalize();
  return 0;
}
